open F

type type_error =
  | TypeMismatch of debruijn_type * debruijn_type
  | ArityMismatch of arity_requirement * debruijn_type
  | NotAnArrow of debruijn_type
  | NotAProduct of debruijn_type
  | NotAForall of debruijn_type
  | NotAnEqual of debruijn_type
  | UnboundTermVariable of tevar
  | UnboundTypeVariable of tyvar
  | TwoOccurencesOfSameVariable of string
  | ContextNotAbsurd

and arity_requirement = Index of int | Total of int

exception TypeError of type_error

(* -------------------------------------------------------------------------- *)

(* A type environment maps term variables to types in de Bruijn's representation. *)

(* Our term variables are in nominal representation, which allows us to represent
   the environment as a binary search tree. One difficulty is that when we enter a
   new type variable binder [TyAbs], we must (conceptually) shift every type in
   the range of the environment up by one. When term variables are in de Bruijn
   notation and the environment is a list, this is easily done by inserting a mark
   in front of the environment. Here, we must be more clever. We maintain an [N2DB]
   environment that tells us, for every *term* variable, how long ago it was bound,
   that is, how many *type* variables were introduced since then. Hence, this tells
   us by how much we must shift its type. It's really a bit too clever, but let's
   do this for fun. *)

module TermVar =
  String

module TermVarMap =
  Map.Make(TermVar)

module N2DB =
  DeBruijn.Nominal2deBruijn(TermVar)

(* We use a union-find data structure to represent equalities between rigid
   variables and types. We need the ability to snapshot and rollback changes, to
   undo the addition of an equality to the environment. *)
module UF = struct
  include UnionFind.Make(UnionFind.StoreVector)
  type 'a elem = 'a rref
end

module S =
  Structure

(* We translate types into union-find graphs to check equality.
   μ-types create cyclic graphs. Each vertex represents a type expression.
   GADTs can add type equalities in the context and we represent
   this by unifying their vertices in the union-find graph. This is
   possible only when they have compatible structures (or one of the
   side has no structure), otherwise the equality is inconsistent. *)
type vertex =
  structure option UF.elem

and structure =
  (* Corresponds to types from the source language (arrow, product, ...). *)
  | SStruct of vertex S.structure
  (* Represents ∀α.τ. We use De Bruijn levels, so we don't need to name
     explicitly the bound variable α. *)
  | SForall of vertex
  (* A De Bruijn level corresponding to a ∀-bound variable. Using levels
     instead of indices makes the equality test easier. *)
  | SForallLevel of int

(* The global store of the union-find. *)
type store = structure option UF.store

(* A mapping from rigid variables to union-find vertices. *)
type tyvars = structure option UF.elem list

(* If the type equations introduced are inconsistent, then no need for
   [store] and [tyvars] *)
type consistency =
  | Consistent of store * tyvars
  | Inconsistent

type env = {
  (* A mapping of term variables to types. *)
  types: debruijn_type TermVarMap.t;
  (* A translation environment of TERM variables to TYPE indices. *)
  names: N2DB.env;
  (* We keep trace of wheter or not the typing equations is consistent *)
  uf_state: consistency
}

let empty = {
  types = TermVarMap.empty;
  names = N2DB.empty;
  uf_state = Consistent (UF.new_store(), []);
}

let lookup { types; names; _ } x =
  try
    (* Obtain the type associated with [x]. *)
    let ty = TermVarMap.find x types in
    (* Find how long ago [x] was bound. *)
    let w = N2DB.lookup names x in
    (* Shift the type [ty] by this amount, so that it makes sense in the
       current scope. *)
    Type.lift w 0 ty
  with Not_found ->
    (* must have been raised by [TermVarMap.find] *)
    raise (TypeError (UnboundTermVariable x))

let extend_with_tevar { types; names; uf_state; } x ty =
  (* Map the name [x] to [ty], and record when it was bound, without
     incrementing the time. *)
  { types = TermVarMap.add x ty types;
    names = N2DB.slide names x;
    uf_state; }

let extend_with_tyvar { types; names; uf_state; } =
  match uf_state with
  | Inconsistent ->
     (* TODO : verify that it is correct *)
     { types; names; uf_state; }
  | Consistent (store, tyvars) ->
     (* Create a fresh vertex for the new type variable *)
     let v = UF.make store None in
     (* Extend the environment of type variables *)
     let tyvars = v :: tyvars in
     (* Increment the time. *)
     let names = N2DB.bump names in
     { types; names ; uf_state = Consistent(store,tyvars); }

let singleton x ty =
  extend_with_tevar empty x ty

let concat env1 env2 =
  let combine _ _ ty2 = Some ty2 in
  let types = TermVarMap.union combine env1.types env2.types in
  let names = N2DB.concat env1.names env2.names in
  match (env1.uf_state, env2.uf_state) with
  | Inconsistent,_ | _, Inconsistent ->
     { types; names; uf_state = Inconsistent }
  | Consistent (store1, tyvars1),
    Consistent (store2, tyvars2) ->
     assert (store1 == store2);
     let store = store1 in
     let tyvars = tyvars1 @ tyvars2 in
     { types; names; uf_state = Consistent (store, tyvars); }

let concat_disjoint env1 env2 =
  let exception Conflict of TermVarMap.key in
  try
    let combine x _ _ =
      raise (Conflict x) in
    let types = TermVarMap.union combine env1.types env2.types in
    let names = N2DB.concat env1.names env2.names in
    match (env1.uf_state, env2.uf_state) with
    | Inconsistent,_ | _, Inconsistent ->
       Ok { types; names; uf_state = Inconsistent }
    | Consistent (store1, tyvars1),
      Consistent (store2, tyvars2) ->
       assert (store1 == store2);
       let store = store1 in
       let tyvars = tyvars1 @ tyvars2 in
       Ok { types; names; uf_state = Consistent (store, tyvars); }
  with Conflict x -> Error x

(* -------------------------------------------------------------------------- *)

(* Destructors. *)

let unfold ty =
  match ty with
  | TyMu ((), body) ->
      Type.subst ty 0 body
  | _ ->
      assert false

let rec as_arrow ty =
  match ty with
  | TyArrow (ty1, ty2) ->
      ty1, ty2
  | TyMu _ ->
      as_arrow (unfold ty)
  | _ ->
      raise (TypeError (NotAnArrow ty))

let rec as_product ty =
  match ty with
  | TyProduct tys ->
      tys
  | TyMu _ ->
      as_product (unfold ty)
  | _ ->
      raise (TypeError (NotAProduct ty))

let rec as_forall ty =
  match ty with
  | TyForall ((), ty) ->
      ty
  | TyMu _ ->
      as_forall (unfold ty)
  | _ ->
      raise (TypeError (NotAForall ty))

(* -------------------------------------------------------------------------- *)

(* A sound approximation of the equality test. *)

(* The System F types produced by an ML type inference algorithm cannot
   contain [TyForall] under [TyMu]. In this restricted setting, deciding
   type equality is relatively easy. We map each type to a graph of
   unification variables (where every variable has rigid structure),
   then we check that the two graphs can be unified. *)

exception Clash =
  Structure.Iter2

(* This function assumes no universal quantification inside a μ-type.
   That is why we need a boolean [inside_mu] during the traversal of the type
   to remember wether or not we are under a μ-quantification *)
let rec translate ~inside_mu store (env : vertex list) (ty : F.debruijn_type)
  : vertex
  = match ty with

  | TyVar x ->
      begin
        (* If the type is closed, then [x] was bound by a quantifier before and
           was added to the environment. *)
        try
          List.nth env x
        with Not_found ->
          raise (TypeError (UnboundTypeVariable x))
      end

  | TyMu ((), ty) ->
      (* Extend the environment with a mapping of this μ-bound variable
         to a fresh vertex. *)
      let v1 = UF.make store None in
      let env = v1 :: env in
      (* Translate the body. *)
      let v2 = translate ~inside_mu:true store env ty in
      (* Unify the vertices [v1] and [v2], keeping [v2]'s structure. *)
      let v = UF.merge store (fun _ s2 -> s2) v1 v2 in
      v

  | TyForall ((), ty) ->
      assert (not inside_mu);
      (* Extend the environment with a mapping of this forall-bound variable
         to a fresh vertex. *)
      let v1 = UF.make store (Some (SForallLevel (List.length env))) in
      let env = v1 :: env in
      (* Translate the body. *)
      let v2 = translate ~inside_mu store env ty in
      UF.make store (Some (SForall v2))

  | TyArrow (ty1, ty2) ->
      let v1 = translate ~inside_mu store env ty1
      and v2 = translate ~inside_mu store env ty2 in
      UF.make store (Some (SStruct (S.TyArrow (v1, v2))))

  | TyProduct tys ->
      let vs = translate_list ~inside_mu store env tys in
      UF.make store (Some (SStruct (S.TyProduct vs)))

  | TyConstr (id, tys) ->
      let vs = translate_list ~inside_mu store env tys in
      UF.make store (Some (SStruct (S.TyConstr (id, vs))))

  | TyEq (ty1, ty2) ->
      let v1 = translate ~inside_mu store env ty1
      and v2 = translate ~inside_mu store env ty2 in
      UF.make store (Some (SStruct (S.TyEq (v1, v2))))

and translate_list ~inside_mu store env tys =
  List.map (translate ~inside_mu store env) tys

let insert q v1 v2 =
  Stack.push (v1, v2) q

let unify_struct q s1 s2 =
  match s1, s2 with
  | SForallLevel x1, SForallLevel x2 ->
      if x1 <> x2 then raise Clash
  | SStruct s1, SStruct s2 ->
      Structure.iter2 (insert q) s1 s2
  | SForall s1, SForall s2 ->
      insert q s1 s2
  | SForallLevel _, _
  | SStruct _, _
  | SForall _, _ ->
      raise Clash

(* The following functions serve two purposes :
   - decompose an equality assumption
   - perform an equality check
   Since these two operations can be break down into similar parts, we
   factorize the code, by passing a [mode] argument to the functions. *)
type mode = Input | Check

(* The [mode] tells us if one should unify rigid variables with structure, and
   add new equalities in the typing context (Input mode), or if we are checking
   for equality and such unifications are not allowed (Check mode). *)
let merge_struct mode q so1 so2 : structure option =
  match so1, so2 with
  | Some s1, Some s2 ->
     unify_struct q s1 s2;
     so1
  | None, so | so, None ->
     match mode with
     | Input ->
        so
     | Check ->
        raise Clash

let rec unify mode store q (v1 : vertex) (v2 : vertex) =
  let (_ : vertex) = UF.merge store (merge_struct mode q) v1 v2 in
  unify_pending mode store q

and unify_pending mode store q =
  match Stack.pop q with
  | v1, v2 ->
      unify mode store q v1 v2
  | exception Stack.Empty ->
      ()

let unify mode store v1 v2 =
  let q = Stack.create() in
  unify mode store q v1 v2

let unify_types mode env ty1 ty2 =
  match env.uf_state with
  | Inconsistent ->
     true
  | Consistent (store, tyvars) ->
     try
       let v1 = translate ~inside_mu:false store tyvars ty1
       and v2 = translate ~inside_mu:false store tyvars ty2 in
       unify mode store v1 v2;
       true
     with Clash ->
       false

(* -------------------------------------------------------------------------- *)

(* Re-package the type equality assumption as a type consistency check. *)
let consistent env ty1 ty2 : bool =
  unify_types Input env ty1 ty2

(* Re-package the type equality test as a type equality check. *)
let equal env ty1 ty2 : unit =
  if not (unify_types Check env ty1 ty2) then
    raise (TypeError (TypeMismatch (ty1, ty2)))

(* -------------------------------------------------------------------------- *)

(* The type-checker. *)

let nth_type ty tys i =
  if i < 0 || i >= List.length tys then
    raise (TypeError (ArityMismatch (Index i, ty)))
  else
    List.nth tys i

let rec typeof datatype_env env (t : debruijn_term) : debruijn_type =
  let typeof = typeof datatype_env in
  match t with
  | Var (_, x) ->
      lookup env x
  | Hole (_, ty, ts) ->
      let on_subterm t =
        ignore (typeof env t) in
      List.iter on_subterm ts;
      ty
  | Annot (_, t, ty) ->
      equal env (typeof env t) ty;
      ty
  | Abs (_, x, ty1, t) ->
      let ty2 = typeof (extend_with_tevar env x ty1) t in
      TyArrow (ty1, ty2)
  | App (_, t, u) ->
      let ty1, ty2 = as_arrow (typeof env t) in
      equal env (typeof env u) ty1;
      ty2
  | Let (_, x, t, u) ->
      let env = extend_with_tevar env x (typeof env t) in
      typeof env u
  | TyAbs (_, (), t) ->
      TyForall ((), typeof (extend_with_tyvar env) t)
  | TyApp (_, t, ty2) ->
      Type.subst ty2 0 (as_forall (typeof env t))
  | Tuple (_, ts) ->
      let tys = List.map (typeof env) ts in
      TyProduct tys
  | Proj (_, i, t) ->
      let ty = typeof env t in
      let tys = as_product ty in
      nth_type ty tys i
  | Variant (_, lbl, dty, t) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      let ty = typeof env t in
      equal env ty arg_type;
      TyConstr dty
  | LetProd (_, xs, t, u) ->
      let ty = typeof env t in
      let tys = as_product ty in
      if List.length xs <> List.length tys then
        raise (TypeError (ArityMismatch (Total (List.length xs), ty)));
      let env = List.fold_left2 extend_with_tevar env xs tys in
      typeof env u
  | Match (_, ty, t, brs) ->
      let scrutinee_ty = typeof env t in
      let tys = List.map (typeof_branch datatype_env env scrutinee_ty) brs in
      List.iter (fun ty2 -> equal env ty ty2) tys;
      ty
  | Absurd (_, ty) ->
      begin
        match env.uf_state with
        | Inconsistent ->
           ty
        | Consistent _ ->
           raise (TypeError (ContextNotAbsurd))
      end
  | Refl (_, ty) ->
      TyEq (ty, ty)
  | Use (_, t, u) ->
      match env.uf_state with
      | Inconsistent ->
         typeof env u
      | Consistent (store, tyvars) ->
         let store_copy = UF.copy store in
         let env = { env with uf_state = Consistent (store_copy, tyvars) } in
         begin
           match typeof env t with
           | TyEq (ty1, ty2) ->
              if consistent env ty1 ty2 then
                (* typecheck [u] in a context containing equalities introduced
                   by the equality ty1 = ty2 *)
                typeof env u
              else
                typeof { env with uf_state = Inconsistent } u
           | ty ->
              raise (TypeError (NotAnEqual ty))
         end

and typeof_variant datatype_env lbl (tid, tyargs) =
  let Datatype.{ data_kind ; type_params ; labels_descr; _ } =
    Datatype.Env.find_decl datatype_env tid in
  let Datatype.{ arg_type; _ } =
    List.find (fun l -> l.Datatype.label_name = lbl) labels_descr in
  begin
    match data_kind with
    | Datatype.Variant ->
        ()
    | Datatype.Record ->
        raise Datatype.Env.UnexpectedRecord
  end;
  List.fold_right2 (fun ty () arg_type ->
    Type.subst ty 0 arg_type
  ) tyargs type_params arg_type

and typeof_branch datatype_env env scrutinee_ty (pat, rhs) =
  let new_bindings = binding_pattern datatype_env env scrutinee_ty pat in
  let new_env = concat env new_bindings in
  typeof datatype_env new_env rhs

and binding_pattern datatype_env env scrutinee_ty pat =
  let binding_pattern = binding_pattern datatype_env env in
  match pat with
  | PVar (_, x) ->
      singleton x scrutinee_ty
  | PWildcard _ ->
      empty
  | PAnnot (_, pat, ty) ->
      equal env scrutinee_ty ty;
      binding_pattern ty pat
  | PTuple (_, ps) ->
      let tys = as_product scrutinee_ty in
      let envs = List.map2 binding_pattern tys ps in
      let f env1 env2 =
        match concat_disjoint env1 env2 with
        | Ok env -> env
        | Error x -> raise (TypeError (TwoOccurencesOfSameVariable x))
      in
      List.fold_left f empty envs
  | PVariant (_, lbl, dty, pat) ->
      let arg_type = typeof_variant datatype_env lbl dty in
      equal env scrutinee_ty (TyConstr dty);
      binding_pattern arg_type pat

let typeof datatype_env t =
  typeof datatype_env empty t

let typeof_result datatype_env u =
  match typeof datatype_env u with
  | v -> Ok v
  | exception TypeError e ->
    Error e
